@@ -64,7 +64,7 @@ G_DEFINE_TYPE (CcCropArea, cc_crop_area, GTK_TYPE_WIDGET);
 static void
 update_image_and_crop (CcCropArea *area)
 {
-    GtkAllocation allocation;
+    int widget_width, widget_height;
     int width, height;
     int dest_width, dest_height;
     double scale;
@@ -72,16 +72,18 @@ update_image_and_crop (CcCropArea *area)
     if (area->paintable == NULL)
         return;
 
-    gtk_widget_get_allocation (GTK_WIDGET (area), &allocation);
+    widget_width = gtk_widget_get_width (GTK_WIDGET (area));
+    widget_height = gtk_widget_get_height (GTK_WIDGET (area));
 
     /* Get the size of the paintable */
     width = gdk_paintable_get_intrinsic_width (area->paintable);
     height = gdk_paintable_get_intrinsic_height (area->paintable);
 
+    if (width == 0 || height == 0)
+        return;
+
     /* Find out the scale to convert to widget width/height */
-    scale = allocation.height / (double) height;
-    if (scale * width > allocation.width)
-        scale = allocation.width / (double) width;
+    scale = MIN (widget_height / (double) height, widget_width / (double) width);
 
     dest_width = width * scale;
     dest_height = height * scale;
@@ -102,8 +104,8 @@ update_image_and_crop (CcCropArea *area)
     }
 
     area->scale = scale;
-    area->image.x = (allocation.width - dest_width) / 2;
-    area->image.y = (allocation.height - dest_height) / 2;
+    area->image.x = (widget_width - dest_width) / 2;
+    area->image.y = (widget_height - dest_height) / 2;
     area->image.width = dest_width;
     area->image.height = dest_height;
 }
@@ -445,73 +447,78 @@ on_drag_cancel (CcCropArea       *area,
     area->drag_offy = 0;
 }
 
-#define CORNER_LINE_WIDTH 4.0
-#define CORNER_LINE_LENGTH 15.0
-#define CORNER_SIZE (CORNER_LINE_LENGTH + CORNER_LINE_WIDTH / 2)
+#define CORNER_LINE_WIDTH  (4.0)
+#define CORNER_LINE_LENGTH (15.0)
+#define CORNER_LINE_COLOR  ((GdkRGBA) { 1, 1, 1, 1 })
 
 static void
 cc_crop_area_snapshot (GtkWidget   *widget,
                        GtkSnapshot *snapshot)
 {
     CcCropArea *area = CC_CROP_AREA (widget);
-    cairo_t *cr;
     GdkRectangle crop;
+    GskRoundedRect crop_circle;
+    graphene_rect_t img_bounds = GRAPHENE_RECT_INIT (0, 0, area->image.width, area->image.height);
 
     if (area->paintable == NULL)
         return;
 
     update_image_and_crop (area);
 
-
     gtk_snapshot_save (snapshot);
 
-    /* First draw the picture */
-    gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (area->image.x, area->image.y));
-
-    gdk_paintable_snapshot (area->paintable, snapshot, area->image.width, area->image.height);
-
-    /* Draw the cropping UI on top with cairo */
-    cr = gtk_snapshot_append_cairo (snapshot, &GRAPHENE_RECT_INIT (0, 0, area->image.width, area->image.height));
-
+    /* Calculate the crop circle */
     get_scaled_crop (area, &crop);
     crop.x -= area->image.x;
     crop.y -= area->image.y;
 
-    /* Draw the circle as an ellipse, to prevent rounding from jitter of the edges */
-    cairo_save (cr);
-    cairo_translate (cr, crop.x + crop.width / 2.0, crop.y + crop.height / 2.0);
-    cairo_scale (cr, crop.width / 2.0, crop.height / 2.0);
-    cairo_arc (cr, 0, 0, 1, 0, 2 * G_PI);
-    cairo_restore (cr);
-    cairo_save (cr);
-    cairo_rectangle (cr, 0, 0, area->image.width, area->image.height);
-    cairo_set_source_rgba (cr, 0, 0, 0, 0.4);
-    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
-    cairo_fill (cr);
-    cairo_restore (cr);
+    gsk_rounded_rect_init_from_rect (&crop_circle,
+                                     &GRAPHENE_RECT_INIT (crop.x, crop.y, crop.width, crop.height),
+                                     crop.width / 2.0);
 
-    /* draw the four corners */
-    cairo_set_source_rgb (cr, 1, 1, 1);
-    cairo_set_line_width (cr, CORNER_LINE_WIDTH);
+    /* Translate to the correct point for rendering the image */
+    gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (area->image.x, area->image.y));
+
+    /* Draw the picture */
+    gdk_paintable_snapshot (area->paintable, snapshot, area->image.width, area->image.height);
+
+    /* Draw inverted circle mask on top */
+    gtk_snapshot_push_mask (snapshot, GSK_MASK_MODE_INVERTED_ALPHA);
+
+    gtk_snapshot_push_rounded_clip (snapshot, &crop_circle);
+    gtk_snapshot_append_color (snapshot, &(GdkRGBA){0, 0, 0, 1}, &img_bounds);
+    gtk_snapshot_pop (snapshot);
+    gtk_snapshot_pop (snapshot);
+
+    gtk_snapshot_push_opacity (snapshot, 0.4);
+    gtk_snapshot_append_color (snapshot, &(GdkRGBA){0, 0, 0, 1}, &img_bounds);
+    gtk_snapshot_pop (snapshot);
+    gtk_snapshot_pop (snapshot);
+
+    /* Draw the corners */
+#define HORIZONTAL_LINE(x, y) \
+    gtk_snapshot_append_color (snapshot, &CORNER_LINE_COLOR, \
+                               &GRAPHENE_RECT_INIT ((x), (y), CORNER_LINE_LENGTH, CORNER_LINE_WIDTH))
+#define VERTICAL_LINE(x, y) \
+    gtk_snapshot_append_color (snapshot, &CORNER_LINE_COLOR, \
+                               &GRAPHENE_RECT_INIT ((x), (y), CORNER_LINE_WIDTH, CORNER_LINE_LENGTH))
 
     /* top left corner */
-    cairo_move_to (cr, crop.x + CORNER_LINE_WIDTH / 2, crop.y + CORNER_SIZE);
-    cairo_rel_line_to (cr, 0, -CORNER_LINE_LENGTH);
-    cairo_rel_line_to (cr, CORNER_LINE_LENGTH, 0);
+    HORIZONTAL_LINE (crop.x, crop.y);
+    VERTICAL_LINE (crop.x, crop.y);
     /* top right corner */
-    cairo_rel_move_to (cr, crop.width - 2 * CORNER_SIZE, 0);
-    cairo_rel_line_to (cr, CORNER_LINE_LENGTH, 0);
-    cairo_rel_line_to (cr, 0, CORNER_LINE_LENGTH);
+    HORIZONTAL_LINE (crop.x + crop.width - CORNER_LINE_LENGTH, crop.y);
+    VERTICAL_LINE (crop.x + crop.width - CORNER_LINE_WIDTH, crop.y);
     /* bottom right corner */
-    cairo_rel_move_to (cr, 0, crop.height - 2 * CORNER_SIZE);
-    cairo_rel_line_to (cr, 0, CORNER_LINE_LENGTH);
-    cairo_rel_line_to (cr, -CORNER_LINE_LENGTH, 0);
+    HORIZONTAL_LINE (crop.x + crop.width - CORNER_LINE_LENGTH,
+                     crop.y + crop.height - CORNER_LINE_WIDTH);
+    VERTICAL_LINE (crop.x + crop.width - CORNER_LINE_WIDTH,
+                   crop.y + crop.height - CORNER_LINE_LENGTH);
     /* bottom left corner */
-    cairo_rel_move_to (cr, -(crop.width - 2 * CORNER_SIZE), 0);
-    cairo_rel_line_to (cr, -CORNER_LINE_LENGTH, 0);
-    cairo_rel_line_to (cr, 0, -CORNER_LINE_LENGTH);
-
-    cairo_stroke (cr);
+    HORIZONTAL_LINE (crop.x,
+                     crop.y + crop.height - CORNER_LINE_WIDTH);
+    VERTICAL_LINE (crop.x,
+                   crop.y + crop.height - CORNER_LINE_LENGTH);
 
     gtk_snapshot_restore (snapshot);
 }
@@ -574,17 +581,18 @@ cc_crop_area_new (void)
 }
 
 /**
- * cc_crop_area_create_pixbuf:
+ * cc_crop_area_create_texture:
  * @area: A crop area
  *
  * Renders the area's paintable, with the cropping applied by the user, into a
- * GdkPixbuf.
+ * GdkTexture.
  *
  * Returns: (transfer full): The cropped picture
  */
-GdkPixbuf *
-cc_crop_area_create_pixbuf (CcCropArea *area)
+GdkTexture *
+cc_crop_area_create_texture (CcCropArea *area)
 {
+    g_autoptr (GdkPaintable) paintable = NULL;
     g_autoptr (GtkSnapshot) snapshot = NULL;
     g_autoptr (GskRenderNode) node = NULL;
     g_autoptr (GskRenderer) renderer = NULL;
@@ -595,9 +603,10 @@ cc_crop_area_create_pixbuf (CcCropArea *area)
     g_return_val_if_fail (CC_IS_CROP_AREA (area), NULL);
 
     snapshot = gtk_snapshot_new ();
-    gdk_paintable_snapshot (area->paintable, snapshot,
-                            gdk_paintable_get_intrinsic_width (area->paintable),
-                            gdk_paintable_get_intrinsic_height (area->paintable));
+    paintable = gdk_paintable_get_current_image (area->paintable);
+    gdk_paintable_snapshot (paintable, snapshot,
+                            gdk_paintable_get_intrinsic_width (paintable),
+                            gdk_paintable_get_intrinsic_height (paintable));
     node = gtk_snapshot_free_to_node (g_steal_pointer (&snapshot));
 
     renderer = gsk_gl_renderer_new ();
@@ -610,7 +619,7 @@ cc_crop_area_create_pixbuf (CcCropArea *area)
     texture = gsk_renderer_render_texture (renderer, node, &viewport);
     gsk_renderer_unrealize (renderer);
 
-    return gdk_pixbuf_get_from_texture (texture);
+    return g_steal_pointer (&texture);
 }
 
 /**
@@ -629,6 +638,15 @@ cc_crop_area_get_paintable (CcCropArea *area)
     return area->paintable;
 }
 
+static void
+on_paintable_invalidated (GdkPaintable *paintable,
+                          gpointer      user_data)
+{
+    GtkWidget *widget = GTK_WIDGET (user_data);
+
+    gtk_widget_queue_draw (widget);
+}
+
 void
 cc_crop_area_set_paintable (CcCropArea   *area,
                             GdkPaintable *paintable)
@@ -636,6 +654,21 @@ cc_crop_area_set_paintable (CcCropArea   *area,
     g_return_if_fail (CC_IS_CROP_AREA (area));
     g_return_if_fail (GDK_IS_PAINTABLE (paintable));
 
+    if (area->paintable) {
+        const guint flags = gdk_paintable_get_flags (paintable);
+
+        if ((flags & GDK_PAINTABLE_STATIC_CONTENTS) == 0)
+            g_signal_handlers_disconnect_by_func (area->paintable,
+                                                  on_paintable_invalidated,
+                                                  area);
+
+        if ((flags & GDK_PAINTABLE_STATIC_SIZE) == 0)
+            g_signal_handlers_disconnect_by_func (area->paintable,
+                                                  on_paintable_invalidated,
+                                                  area);
+
+    }
+
     g_set_object (&area->paintable, paintable);
 
     area->scale = 0.0;
@@ -644,7 +677,25 @@ cc_crop_area_set_paintable (CcCropArea   *area,
     area->image.width = 0;
     area->image.height = 0;
 
+    if (paintable) {
+        const guint flags = gdk_paintable_get_flags (paintable);
+
+        if ((flags & GDK_PAINTABLE_STATIC_CONTENTS) == 0)
+          g_signal_connect (paintable,
+                            "invalidate-contents",
+                            G_CALLBACK (on_paintable_invalidated),
+                            area);
+
+        if ((flags & GDK_PAINTABLE_STATIC_SIZE) == 0)
+          g_signal_connect (paintable,
+                            "invalidate-size",
+                            G_CALLBACK (on_paintable_invalidated),
+                            area);
+    }
+
+
     gtk_widget_queue_draw (GTK_WIDGET (area));
+    gtk_widget_queue_resize (GTK_WIDGET (area));
 }
 
 /**
